import java.io.*;
import java.util.*;

import java.io.IOException;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

import gov.nih.nci.evs.domain.*;

import java.io.PrintWriter;

import gov.nih.nci.common.util.*;
import gov.nih.nci.evs.domain.*;
import gov.nih.nci.evs.query.*;
import gov.nih.nci.system.applicationservice.*;


public class FDAReportWriter
{
    static final int FDA_CONCEPT_NAME = 0;
    static final int FDA_CONCEPT_CODE = 1;
    static final int FDA_OID_VALUE = 2;
    static final int FDA_TABLE_VALUE = 3;
    static final int FDA_DEFINITION = 4;

    static final String CONCEPT_NAME_STR = "Concept Name";
    static final String CONCEPT_CODE_STR = "Concept Code";
    static final String OID_STR = "OID";
    static final String FDA_TABLE_STR = "FDA_Table";
    static final String FULL_SYN_STR = "FULL_SYN";
    static final String DEFINITION_STR = "DEFINITION";

    static final String targetpropertyname = "FDA_Table";


    private String[] fda_data = new String[5];
    private Vector name_vec;

    HashSet hset;

	private PrintWriter out;
    Hashtable hyperprops;
    String displayname;
    String httpURL;
    int CharsPerLine = 80;
    ApplicationService appService;
    EVSQuery evsQuery;

    String vocabulary;



    Vector attribute_name_vec;
    Vector attribute_value_vec;

    String asciifile;

	public FDAReportWriter(String httpURL) {
		this.httpURL = httpURL;
	}

    public boolean setVocabulary(String vocabulary)
    {
		this.vocabulary = vocabulary;
		return(initialize());
	}

	private void reset_data()
	{
		for (int i=0; i<fda_data.length; i++)
		{
			fda_data[i] = "";
		}
		name_vec.clear();
	}

	private boolean initialize() {
		try{
			appService = ApplicationService.getRemoteInstance(httpURL);
			evsQuery = new EVSQueryImpl();
			if (appService == null)
			{
				System.out.println("Unable to create ApplicationService instance -- program aborts.");
				System.exit(1);
			}
			System.out.println("ApplicationService instance created.");

            asciifile = "FDA_Table.txt";
		    out = new PrintWriter(new BufferedWriter(new FileWriter(asciifile)));
		    printLabels();

		    attribute_name_vec = new Vector();
		    attribute_value_vec = new Vector();

            name_vec = new Vector();
            reset_data();
            hset = new HashSet();



            return true;
		} catch(Exception e) {
			System.err.println(e);
			e.printStackTrace();
			return false;
		}
	}


    boolean reported(String conceptname)
    {
		if (hset.contains(conceptname))
		{
			return true;
		}
		hset.add(conceptname);
		return false;
	}

    private Vector arrayToVector(Object[] objects)
    {
		Vector v = new Vector();
		if (objects == null)
			return v;

		for (int i=0; i<objects.length; i++)
			v.add(objects[i]);
		return v;
    }

	private Object[] getRootConcepts(String vocabularyName, boolean allAttributes)
	{
		Object[] ret_obj_array = null;
		try
		{
			EVSQuery evsQuery = new EVSQueryImpl();
			evsQuery.getRootConcepts(vocabularyName, allAttributes);
			List evsResults = new ArrayList();
			evsResults = (List)appService.evsSearch(evsQuery);
			if (evsResults == null || evsResults.size() == 0) return null;
			return (Object[]) evsResults.toArray();
		}
		catch(Exception e){
			e.printStackTrace();
		}
		return ret_obj_array;
	}


	private DescLogicConcept getDescLogicConcept(String vocabularyName, String conceptName, boolean inputFlag)
	{
		DescLogicConcept ret_DescLogicConcept = null;
		try
		{
			EVSQuery evsQuery = new EVSQueryImpl();
			evsQuery.getDescLogicConcept(vocabularyName, conceptName, inputFlag);
			List evsResults = new ArrayList();
			evsResults = (List)appService.evsSearch(evsQuery);
			if (evsResults == null || evsResults.size() == 0) return null;
			Object[] objs = evsResults.toArray();
			ret_DescLogicConcept = (DescLogicConcept) objs[0];
		}
		catch(Exception e){
			e.printStackTrace();
		}
		return ret_DescLogicConcept;
	}



	private Vector getSubConcepts(String vocabularyName, String conceptName, java.lang.Boolean inputFlag, java.lang.Boolean outputFlag)
	{
		Vector ret_Vector = null;
		try
		{
			EVSQuery evsQuery = new EVSQueryImpl();
			evsQuery.getSubConcepts(vocabularyName, conceptName, inputFlag, outputFlag);
			List evsResults = new ArrayList();
			evsResults = (List)appService.evsSearch(evsQuery);
			if (evsResults == null || evsResults.size() == 0) return null;
			return arrayToVector(evsResults.toArray());
		}
		catch(Exception e){
			e.printStackTrace();
		}
		return ret_Vector;
	}


	private Vector getPropertyValues(String vocabularyName, String conceptName, String propertyName)
	{
		Vector ret_Vector = null;
		try
		{
			EVSQuery evsQuery = new EVSQueryImpl();
			evsQuery.getPropertyValues(vocabularyName, conceptName, propertyName);
			List evsResults = new ArrayList();
			evsResults = (List)appService.evsSearch(evsQuery);
			if (evsResults == null || evsResults.size() == 0) return null;
			return arrayToVector(evsResults.toArray());
		}
		catch(Exception e){
			e.printStackTrace();
		}
		return ret_Vector;
	}


	private DescLogicConcept[] searchDescLogicConcepts(String vocabularyName,
	                                                  String searchTerm,
	                                                  int limit,
	                                                  int matchOption,
	                                                  String matchType,
	                                                  int ASDIndex)
	{
		DescLogicConcept[] ret_DescLogicConcept_array = null;
		try
		{
			EVSQuery evsQuery = new EVSQueryImpl();
			evsQuery.searchDescLogicConcepts(vocabularyName, searchTerm, limit, matchOption, matchType, ASDIndex);
			List evsResults = new ArrayList();
			evsResults = (List)appService.evsSearch(evsQuery);
			if (evsResults == null || evsResults.size() == 0) return null;
			Object[] objs = evsResults.toArray();
			ret_DescLogicConcept_array = new DescLogicConcept[objs.length];
			for (int i=0; i<objs.length; i++)
			{
				ret_DescLogicConcept_array[i] = (DescLogicConcept) objs[i];
			}
		}
		catch(Exception e){
			e.printStackTrace();
		}
		return ret_DescLogicConcept_array;
	}


	private Vector findFDARootConcepts(String oid_value)
	{
		 Vector v = new Vector();
		 DescLogicConcept[] concepts = searchDescLogicConcepts(vocabulary, oid_value, 100, 2, OID_STR, 1);

		 if (concepts == null)
		 {
			 System.out.println("searchDescLogicConcepts returns null");
			 System.exit(1);
		 }
		 for (int i=0; i<concepts.length; i++)
		 {
			 DescLogicConcept c = (DescLogicConcept) concepts[i];
			 System.out.println("Checking " + c.getName() + "...");
			 Vector w = getPropertyValues(vocabulary, c.getName(), OID_STR);
             if (w != null && w.size() > 0)
             {
				 v.add(c);
			     System.out.println("Adding " + c.getName() + " to the roots.");
			 }
			 else
			 {
			     System.out.println("WARNING: " + c.getName() + " does not contains any OID property.");
			 }
		 }
		 return v;
	}

	private void report(Vector conceptnames)
	{
		try {
			for (int i=0; i<conceptnames.size(); i++)
			{
				String conceptname = (String) conceptnames.elementAt(i);
				System.out.println("");
				System.out.println("Root: " + conceptname);
				report(conceptname);
				reportSubconcepts(conceptname);
			}
	    }
		catch(Exception e) {
			e.printStackTrace();
		}

		closePrintWriter();
	}

	private void reportSubconcepts(String conceptname)
	{
		Vector subs = getSubConcepts(vocabulary, conceptname, Boolean.FALSE, Boolean.FALSE);

		if (subs == null) return;

		for (int i=0; i<subs.size(); i++)
		{
			String sub = (String) subs.elementAt(i);
			System.out.println(conceptname + "|" + sub);
			report(sub, true);

			reportSubconcepts(sub);
		}
	}

	private boolean containsProperty(Vector properties, String propertyname)
	{
		if (properties == null) return false;
		for (int i=0; i<properties.size(); i++)
		{
			Property prop = (Property) properties.elementAt(i);
			if (prop.getName().equals(propertyname))
			{
				return true;
			}
		}
		return false;
	}

	private boolean report(String conceptname)
	{
		return report(conceptname, false);
	}


	private boolean report(String conceptname, boolean checkProperty)
	{
		if (reported(conceptname))
		{
			System.out.println(conceptname + " has already been reported.");
			return false;
		}

		try {
			DescLogicConcept concept = null;
			concept = getDescLogicConcept(vocabulary, conceptname, false);

			if (concept == null)
			{
				out.println(conceptname + " not found in vocabulary " + vocabulary);
				out.println("");
				return false;
			}

			Vector properties = concept.getPropertyCollection();
			if (properties != null)
			{
				Property[] property_array = new Property[properties.size()];
				for (int k=0; k<properties.size(); k++)
				{
					property_array[k] = (Property) properties.elementAt(k);
				}

				property_array = sort_property (property_array);

				properties.removeAllElements();
				for (int k=0; k<property_array.length; k++)
				{
					properties.add((Property) property_array[k]);
				}
			}

			if (checkProperty && !containsProperty(properties, targetpropertyname))
			{
				return false;
			}

			logAttribute(CONCEPT_NAME_STR,  conceptname);

			String code = concept.getCode();

			if (code.compareTo("") != 0)
			{
				logAttribute(CONCEPT_CODE_STR,  code);
			}

		    Vector names = getPropertyNames (properties);
			for (int i=0; i<names.size(); i++)
			{
				String name = (String) names.elementAt(i);
				Vector values = getPropertyValues(properties, name);
				for (int j=0; j<values.size(); j++)
				{
				    String value = (String) values.elementAt(j);
					value = XML2Pipe(value);
					logAttribute(name, value);
				}
			}
			dumpToASCII();

		} catch(Exception e) {
			e.printStackTrace();
			return false;
		}
		return true;
	}

	private void closePrintWriter()
	{
		if (out != null)
		{
			out.close();
		}
    }

	private String getDisplayname(String conceptname)
	{
		 String s="";
		 try{
			if (displayname==null || displayname.equals("")){
			  s = conceptname;
			}else{
			  Vector propsValue = getPropertyValues(vocabulary, conceptname, displayname);

			  if (propsValue==null)
				s = conceptname;
			  else
				s = (String) propsValue.elementAt(0);
			}
		 }catch(Exception e){
			 e.printStackTrace();
		 }
		 return s;
	}


    private String XML2Pipe(String pattern) {
        int n = pattern.indexOf(">");

		if (n == -1) return pattern;

        String retstr = "";
        int lcv = 0;
        while (n != -1)
        {
			String tag = pattern.substring(0, n+1);
			pattern = pattern.substring(n+1, pattern.length());
			n = pattern.indexOf("<");
			if (n == -1) break;

			String value = pattern.substring(0, n);
			retstr = retstr + value;
            retstr = retstr + "|";
            lcv++;

			pattern = pattern.substring(n, pattern.length());

			n = pattern.indexOf(">");
			if (n == -1) break;

			tag = pattern.substring(0, n+1);

            if (pattern.length() > tag.length())
            {
				pattern = pattern.substring(n+1, pattern.length());
		    }
		    n = pattern.indexOf(">");
	    }

	    if (lcv > 0)
	    {
			retstr = retstr.substring(0, retstr.length()-1);
		}
		else
		{
			retstr = pattern;
		}
		return(retstr);

	}


    private Vector getPropertyValues(Vector properties, String propertyname)

    {
		if (properties == null) return null;
		Vector v = new Vector();
		for (int i=0; i<properties.size(); i++)
		{
			Property property = (Property) properties.elementAt(i);
			if (property.getName().equals(propertyname))
			{
				v.add(property.getValue());
			}
		}
		return v;
    }


    private String replaceChar(String s, int position, char c)
    {
		if (s == null) return s;
		if (position > s.length() - 1) return s;
		char[] a = s.toCharArray();
		a[position] = c;
		String t = "";
		for (int i=0; i<a.length; i++)
		{
			t = t + Character.toString(a[i]);
		}
		return t;
	}


    private Vector getPropertyNames (Vector properties) {
        if (properties == null || properties.equals(null)) {
            return null;
        }
        Vector v = new Vector();
        String prev_name = "";
		for (int i=0; i<properties.size(); i++)
		{
		    Property p = (Property) properties.elementAt(i);
			if (p.getName().compareTo(prev_name) != 0)
			{
				v.add(p.getName());
                prev_name = p.getName();
			}
		}
		return v;
	}

    static Property[] sort_property (Property[] properties) {
        boolean done = false;
        if (properties == null || properties.equals(null)) {
            return properties;
        }
        while (!done) {
            done = true;
            for (int i = 0;i < properties.length-1;i++) {
                if (properties[i].getName().compareTo(properties[i+1].getName())<= 0) {
                } else {
                    done = false;
                    Property tmp = properties[i];
                    properties[i] = properties[i+1];
                    properties[i+1] = tmp;
                }
            }
        }
        return properties;
    }

	private String extractFDAName(String full_syn)
	{
		int n = full_syn.indexOf("|PT|FDA");
		if (n == -1) return "";
		return full_syn.substring(0, n);
	}

	private String extractFDADefinition(String definition)
	{
		int n = definition.indexOf("|");
		if (n == -1) return "";
		if (definition.substring(0, n).compareTo("FDA") == 0)
		{
			return definition.substring(n+1, definition.length());
		}
		return "";
	}


	private void logAttribute(String name0, String value0)
	{
        String name = name0;
        String value = value0;

		name.replace('\n', ' ');
		value.replace('\n', ' ');

        if (name.compareTo(CONCEPT_CODE_STR) == 0)
        {
			fda_data[FDA_CONCEPT_CODE] = value;
		}
        else if (name.compareTo(OID_STR) == 0)
        {
			fda_data[FDA_OID_VALUE] = value;
		}
        else if (name.compareTo(FDA_TABLE_STR) == 0)
        {
			fda_data[FDA_TABLE_VALUE] = value;
		}
        else if (name.compareTo(FULL_SYN_STR) == 0)
        {
			String fda_name = extractFDAName(value);
			if (fda_name.compareTo("") != 0)
			{
				name_vec.add(fda_name);
			}
		}
        else if (name.compareTo(DEFINITION_STR) == 0)
        {
			String fda_definition = extractFDADefinition(value);
			if (fda_definition.compareTo("") != 0)
			{
				fda_data[FDA_DEFINITION] = fda_definition;
			}
		}
	}

	private void printLabels()
	{
		out.println("FDA Preferred Name$Concept Code$OID$FDA_Table$Definition");
	}


	private void dumpToASCII()
	{
        //String s = "";
        for (int i=0; i<name_vec.size(); i++)
        {
			String line = (String) name_vec.elementAt(i);
			for (int j=1; j<fda_data.length; j++)
			{
				String t = (String) fda_data[j];
				line = line + "$";
				line = line + t;
			}
			out.println(line);
	    }
		reset_data();
	}


    public void generateReport(String oid_value)
    {
        Vector v = new Vector();
		Vector u = new Vector();
		System.out.println(OID_STR + ": " + oid_value);

		v = findFDARootConcepts(oid_value);

		if (v.size() == 0)
		{
			System.out.println("No root concept contains the specified OID property.");
			System.exit(1);
		}
		else
		{
			for (int j=0; j<v.size(); j++)
			{
				DescLogicConcept c = (DescLogicConcept) v.elementAt(j);
				u.add(c.getName());
			}
			report(u);
		    System.out.println("Output files " + asciifile + " generated.");
		}
		hset.clear();
    }


    public static void main(String[] args)
    {
        if (args.length != 2 && args.length != 4)
		{
			System.out.println("Usage: java FDAReportWriter <httpURL> <vocabulary> ");
			System.out.println("   or: java FDAReportWriter <httpURL> <vocabulary> -OID <OID value>");
			System.exit(1);
		}

        String httpURL = args[0];
        String vocabulary = args[1];

		FDAReportWriter test = new FDAReportWriter(httpURL);
		boolean retval = test.setVocabulary(vocabulary);

		if (!retval)
		{
			System.out.println("Unable to connect to caCORE server -- program aborts.");
			System.exit(1);
		}

		int n = -1;
		String oid_value = "*";
		for (int i=0; i<args.length; i++)
		{
			if (args[i].compareToIgnoreCase("-OID") == 0)
			{
				n = i;
				if (i+1 < args.length)
				{
					oid_value = args[i+1];
					break;
				}
				else
				{
					System.out.println("Usage: java FDAReportWriter <httpURL> <vocabulary> ");
					System.out.println("   or: java FDAReportWriter <httpURL> <vocabulary> -OID <OID value>");
					System.exit(1);
				}
			}
		}

		test.generateReport(oid_value);
	}

}




